# noqa: D100
__doc__ = """
.. click:: bfb_delivery.cli.split_chunked_route:main
   :prog: split_chunked_route
   :nested: full
"""

import logging

import click
from typeguard import typechecked

from bfb_delivery import split_chunked_route
from bfb_delivery.lib.constants import Defaults, DocStrings

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


@click.command(help=DocStrings.SPLIT_CHUNKED_ROUTE.cli_docstring)
@click.option(
    "--input_path",
    type=str,
    required=True,
    help=DocStrings.SPLIT_CHUNKED_ROUTE.args["input_path"],
)
@click.option(
    "--output_dir",
    type=str,
    required=False,
    default=Defaults.SPLIT_CHUNKED_ROUTE["output_dir"],
    help=DocStrings.SPLIT_CHUNKED_ROUTE.args["output_dir"],
)
@click.option(
    "--output_filename",
    type=str,
    required=False,
    default=Defaults.SPLIT_CHUNKED_ROUTE["output_filename"],
    help=DocStrings.SPLIT_CHUNKED_ROUTE.args["output_filename"],
)
@click.option(
    "--n_books",
    type=int,
    required=False,
    default=Defaults.SPLIT_CHUNKED_ROUTE["n_books"],
    help=DocStrings.SPLIT_CHUNKED_ROUTE.args["n_books"],
)
@click.option(
    "--book_one_drivers_file",
    type=str,
    required=False,
    default=Defaults.SPLIT_CHUNKED_ROUTE["book_one_drivers_file"],
    help=DocStrings.SPLIT_CHUNKED_ROUTE.args["book_one_drivers_file"],
)
@click.option(
    "--date",
    type=str,
    required=False,
    default=Defaults.SPLIT_CHUNKED_ROUTE["date"],
    help=DocStrings.SPLIT_CHUNKED_ROUTE.args["date"],
)
@typechecked
def main(  # noqa: D103
    input_path: str,
    output_dir: str,
    output_filename: str,
    n_books: int,
    book_one_drivers_file: str,
    date: str,
) -> list[str]:
    paths = split_chunked_route(
        input_path=input_path,
        output_dir=output_dir,
        output_filename=output_filename,
        n_books=n_books,
        book_one_drivers_file=book_one_drivers_file,
        date=date,
    )
    return_paths = [str(path.resolve()) for path in paths]
    logger.info(f"Split workbook(s) saved to:\n{return_paths}")

    return return_paths



"""Module for custom exceptions."""


class BFBError(Exception):
    """Base class for exceptions in this module."""

    pass


class AssignmentOutOfRange(BFBError):
    """Raised when a user selects a driver index out of range."""

    pass


class InactiveDriverAssignment(BFBError):
    """Raised when a user selects an inactive driver."""

    pass



"""Utility functions for the dispatch module."""

import logging
import os
from os import getcwd as os_getcwd  # For test patching.
from pathlib import Path
from typing import Any

import requests
from dotenv import load_dotenv
from typeguard import typechecked

from bfb_delivery.lib.dispatch import api_callers

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


@typechecked
def get_circuit_key() -> str:
    """Get the Circuit API key."""
    load_dotenv(Path(os_getcwd()) / ".env")
    key = os.getenv("CIRCUIT_API_KEY")
    if not key:
        raise ValueError(
            "Circuit API key not found. Set the CIRCUIT_API_KEY environment variable."
        )

    return key


# TODO: Pass params instead of forming URL first. ("params", not "json")
# (Would need to then grab params URL for next page, or just add nextpage to params?)
@typechecked
def get_responses(url: str) -> list[dict[str, Any]]:
    """Get all responses from a paginated API endpoint."""
    # Calling the token salsa to trick bandit into ignoring what looks like a hardcoded token.
    next_page_salsa = ""
    next_page_cookie = ""
    responses = []

    while next_page_salsa is not None:
        paged_response_getter = api_callers.PagedResponseGetter(
            page_url=url + str(next_page_cookie)
        )
        paged_response_getter.call_api()

        stops = paged_response_getter._response.json()
        responses.append(stops)
        next_page_salsa = paged_response_getter.next_page_salsa

        if next_page_salsa:
            salsa_prefix = "?" if "?" not in url else "&"
            next_page_cookie = f"{salsa_prefix}pageToken={next_page_salsa}"

    return responses


@typechecked
def get_response_dict(response: requests.Response) -> dict[str, Any]:
    """Safely handle a response that may not be JSON."""
    try:
        response_dict: dict = response.json()
    except Exception as e:
        response_dict = {
            "reason": response.reason,
            "additional_notes": "No-JSON response.",
            "No-JSON response exception:": str(e),
        }
    return response_dict


@typechecked
def concat_response_pages(
    page_list: list[dict[str, Any]], data_key: str
) -> list[dict[str, Any]]:
    """Extract and concatenate the data lists from response pages."""
    data_list = []
    for page in page_list:
        data_list += page[data_key]

    return data_list



"""Utility functions for the formatting module."""

import configparser
import logging
import os
from pathlib import Path

import pandas as pd
from typeguard import typechecked

from bfb_delivery.lib.constants import BookOneDrivers, Columns, ExtraNotes

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


@typechecked
def get_book_one_drivers(file_path: str) -> list[str]:
    """Get the drivers from the book-one driver's file, or the constant if no path.

    Args:
        file_path: Path to the book-one driver's file. If empty, uses a constant list.

    Returns:
        The drivers to include in book one of split chunked routes.
    """
    sheet_one_drivers = [d.value for d in BookOneDrivers]
    if file_path:
        sheet_one_drivers = pd.read_csv(file_path)[Columns.DRIVER].astype(dtype=str).tolist()

    return sheet_one_drivers


@typechecked
def get_extra_notes(file_path: str) -> pd.DataFrame:
    """Get the extra notes from the file, or the constant if no path.

    Args:
        file_path: Path to the extra notes file. If empty, uses a constant DataFrame.

    Returns:
        The extra notes to include in the combined routes.
    """
    extra_notes_df: pd.DataFrame
    if file_path:
        extra_notes_df = pd.read_csv(file_path)
    else:
        extra_notes = ExtraNotes()
        extra_notes_df = extra_notes.df

    validation_sr = extra_notes_df["tag"]
    validation_sr = validation_sr.apply(lambda x: str(x).replace("*", "").strip())
    duplicated_tags = validation_sr[validation_sr.duplicated()].to_list()
    if duplicated_tags:
        raise ValueError(f"Extra notes has duplicated tags: {duplicated_tags}")

    return extra_notes_df


@typechecked
def get_phone_number(key: str, config_path: str = "config.ini") -> str:
    """Get the phone number from the config file.

    Args:
        key: The key in the config file.
        config_path: The path to the config file.

    Returns:
        The phone number.
    """
    section_key = "phone_numbers"
    phone_number = (
        "NO PHONE NUMBER. "
        "See warning in logs for instructions on setting up your config file."
    )
    full_config_path = Path(config_path)
    full_config_path = full_config_path.resolve()
    config_instructions = (
        f"In config file, under '[{section_key}]', add '{key} = (555) 555-5555'."
    )

    if os.path.exists(full_config_path):
        config = configparser.ConfigParser()
        try:
            config.read(full_config_path)
            phone_number = config[section_key][key]
        except KeyError:
            logger.warning(
                f"{key} not found in config file: {full_config_path}. {config_instructions}"
            )
    else:
        logger.warning(
            (
                f"Config file not found: {full_config_path}. "
                f"Create the file. {config_instructions}"
            )
        )

    return str(phone_number)


@typechecked
def map_columns(df: pd.DataFrame, column_name_map: dict[str, str], invert_map: bool) -> None:
    """Map column names in a DataFrame.

    Operates in place.

    Args:
        df: The DataFrame to map.
        column_name_map: The mapping of column names.
        invert_map: Whether to invert the mapping.
    """
    if invert_map:
        column_name_map = {v: k for k, v in column_name_map.items()}

    df.rename(columns=column_name_map, inplace=True)



"""Data cleaning utilities."""

import logging
import warnings
from collections.abc import Callable
from logging import info

import email_validator
import pandas as pd
import phonenumbers
from typeguard import typechecked

from bfb_delivery.lib.constants import MAX_ORDER_COUNT, Columns

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


@typechecked
def format_column_names(columns: list[str]) -> list[str]:
    """Clean column names.

    Just strips whitespace for now.

    Args:
        columns: The column names to clean.

    Returns:
        The cleaned column names.
    """
    columns = [column.strip() for column in columns]
    columns = [column.title() for column in columns]
    # TODO: Other column cleaning? (e.g., remove special characters, set casing)
    # TODO: Validate? Use general constant list? (Or, just use Pandera?)
    # TODO: Make column names StrEnum? Or just make sure they are in the constants list?
    return columns


@typechecked
def format_and_validate_data(df: pd.DataFrame, columns: list[str]) -> None:
    """Clean, format, and validate selected columns in a DataFrame.

    Operates in place.

    Args:
        df: The DataFrame to clean.
        columns: The columns to clean.

    Returns:
        None

    Raises:
        ValueError: If columns are not found in the DataFrame.
        ValueError: If no formatter is found for a column.
    """
    missing_columns = sorted(list(set(columns) - set(df.columns)))
    if missing_columns:
        raise ValueError(f"Columns not found in DataFrame: {missing_columns}.")

    # TODO: Pre-Validate:
    # ints actually integers and not something that gets cast to an int

    # TODO: FutureWarning: Setting an item of incompatible dtype is deprecated and will
    # raise an error in a future version of pandas. Value '' has dtype incompatible with
    # float64, please explicitly cast to a compatible dtype first.
    # Cast all to string to start? Then convert "nan" to ""?
    with warnings.catch_warnings():
        warnings.simplefilter(action="ignore", category=FutureWarning)
        df.fillna("", inplace=True)

    # Could use generic, class, or Pandera? But, this works, and is flexible and transparent.
    formatters_dict = {
        Columns.ADDRESS: _format_and_validate_address,
        Columns.BOX_TYPE: _format_and_validate_box_type,
        Columns.EMAIL: _format_and_validate_email,
        Columns.DRIVER: _format_and_validate_driver,
        Columns.NAME: _format_and_validate_name,
        Columns.NEIGHBORHOOD: _format_and_validate_neighborhood,
        Columns.NOTES: _format_notes,
        Columns.ORDER_COUNT: _format_and_validate_order_count,
        Columns.PHONE: _format_and_validate_phone,
        Columns.PRODUCT_TYPE: _format_and_validate_product_type,
        Columns.STOP_NO: _format_and_validate_stop_no,
    }
    for column in columns:
        formatter_fx: Callable
        try:
            formatter_fx = formatters_dict[column]
        except KeyError as e:
            raise ValueError(f"No formatter found for column: {column}.") from e
        formatter_fx(df=df)

    # TODO: Some common (post-formatting) validations for all columns:
    # Are the prescribed types.
    # Have no nulls (where appropriate).

    return


@typechecked
def _format_and_validate_address(df: pd.DataFrame) -> None:
    """Format the address column."""
    # Avoid modifying values until we need to. Mostly established values used in Circuit.
    # Will hold off on validation/formatting until we've swallowed more of the process
    # and are starting to map etc.
    _format_string(df=df, column=Columns.ADDRESS)
    _validate_col_not_empty(df=df, column=Columns.ADDRESS)
    return


@typechecked
def _format_and_validate_box_type(df: pd.DataFrame) -> None:
    """Format the box type column."""
    _format_and_validate_product_or_box_type(df=df, column=Columns.BOX_TYPE)
    return


@typechecked
def _format_and_validate_product_type(df: pd.DataFrame) -> None:
    """Format the box type column."""
    _format_and_validate_product_or_box_type(df=df, column=Columns.PRODUCT_TYPE)
    return


@typechecked
def _format_and_validate_product_or_box_type(df: pd.DataFrame, column: str) -> None:
    """Format the box type column."""
    _format_and_validate_names_to_upper(df=df, column=column)
    # TODO: What about multiple box types for one stop?
    # Split and format each value separately, then rejoin?
    # TODO: Validate: make enum.StrEnum?
    return


@typechecked
def _format_and_validate_driver(df: pd.DataFrame) -> None:
    """Format the driver column."""
    _format_and_validate_names_title(df=df, column=Columns.DRIVER)
    return


@typechecked
def _format_and_validate_email(df: pd.DataFrame) -> None:
    """Format and validate the email column."""
    _format_string(df=df, column=Columns.EMAIL)

    formatted_emails = []
    invalid_emails = []
    for email in df[Columns.EMAIL]:
        formatted_email = email

        try:
            if email:
                email_info = email_validator.validate_email(email, check_deliverability=False)
                formatted_email = email_info.normalized
        except email_validator.EmailNotValidError as e:
            invalid_emails.append(email)
            logger.warning(f"Invalid email address, {email}: {e}")
            info("Checking for more invalid addresses before raising error.")

        formatted_emails.append(formatted_email)

    if invalid_emails:
        logger.warning(f"Invalid email addresses found:\n{invalid_emails}")
    else:
        df[Columns.EMAIL] = formatted_emails

    return


@typechecked
def _format_and_validate_name(df: pd.DataFrame) -> None:
    """Format the name column."""
    _format_and_validate_names_base(df=df, column=Columns.NAME)
    return


@typechecked
def _format_and_validate_neighborhood(df: pd.DataFrame) -> None:
    """Format the neighborhood column."""
    _format_and_validate_names_to_upper(df=df, column=Columns.NEIGHBORHOOD)
    # TODO: Validate: make enum.StrEnum?
    return


@typechecked
def _format_notes(df: pd.DataFrame) -> None:
    """Format the notes column."""
    _format_string(df=df, column=Columns.NOTES)
    return


@typechecked
def _format_and_validate_order_count(df: pd.DataFrame) -> None:
    """Format the order count column."""
    _format_int(df=df, column=Columns.ORDER_COUNT)
    _validate_order_count(df=df)
    return


@typechecked
def _format_and_validate_phone(df: pd.DataFrame) -> None:
    """Format and validate the phone column."""
    _format_string(df=df, column=Columns.PHONE)
    df[Columns.PHONE] = df[Columns.PHONE].apply(lambda x: x[:-2] if x.endswith(".0") else x)

    formatting_df = df.copy()
    formatting_df["formatted_numbers"] = formatting_df[Columns.PHONE].apply(
        lambda number: "+" + number if (len(number) > 0 and number[0] != "+") else number
    )
    formatting_df["formatted_numbers"] = [
        phonenumbers.parse(number) if len(number) > 0 else number
        for number in formatting_df["formatted_numbers"].to_list()
    ]

    formatting_df["is_valid"] = formatting_df["formatted_numbers"].apply(
        lambda number: (
            phonenumbers.is_valid_number(number)
            if isinstance(number, phonenumbers.phonenumber.PhoneNumber)
            else True
        )
    )
    if not formatting_df["is_valid"].all():
        invalid_numbers = formatting_df[~formatting_df["is_valid"]]
        logger.warning(
            f"Invalid phone numbers found:\n{invalid_numbers[df.columns.to_list()]}"
        )

    # TODO: Use phonenumbers.format_by_pattern to achieve (555) 555-5555 if desired.
    formatting_df["formatted_numbers"] = [
        (
            str(
                phonenumbers.format_number(
                    number, num_format=phonenumbers.PhoneNumberFormat.INTERNATIONAL
                )
            )
            if isinstance(number, phonenumbers.phonenumber.PhoneNumber)
            else number
        )
        for number in formatting_df["formatted_numbers"].to_list()
    ]

    df[Columns.PHONE] = formatting_df["formatted_numbers"]

    return


@typechecked
def _format_and_validate_stop_no(df: pd.DataFrame) -> None:
    """Format the stop number column."""
    _format_int(df=df, column=Columns.STOP_NO)
    _validate_stop_no(df=df)
    return


@typechecked
def _format_and_validate_names_to_upper(df: pd.DataFrame, column: str) -> None:
    """Format a column with names."""
    _format_and_validate_names_base(df=df, column=column)
    df[column] = df[column].apply(lambda name: name.upper())
    return


@typechecked
def _format_and_validate_names_title(df: pd.DataFrame, column: str) -> None:
    """Format a column with names."""
    _format_and_validate_names_base(df=df, column=column)
    df[column] = df[column].apply(lambda name: name.title())
    return


@typechecked
def _format_and_validate_names_base(df: pd.DataFrame, column: str) -> None:
    """Format a column with names."""
    _format_string(df=df, column=column)
    _validate_col_not_empty(df=df, column=column)
    return


@typechecked
def _format_int(df: pd.DataFrame, column: str) -> None:
    """Basic formatting for an integer column."""
    df[column] = df[column].astype(str).str.strip()
    df[column] = df[column].astype(float).astype(int)
    return


@typechecked
def _format_string(df: pd.DataFrame, column: str) -> None:
    """Basic formatting for a string column. Note: Casts to string."""
    df[column] = df[column].astype(str).str.strip()
    return


@typechecked
def _validate_order_count(df: pd.DataFrame) -> None:
    """Validate the order count column."""
    _validate_col_not_empty(df=df, column=Columns.ORDER_COUNT)
    _validate_greater_than_zero(df=df, column=Columns.ORDER_COUNT)

    too_many_orders_df = df[df[Columns.ORDER_COUNT] > MAX_ORDER_COUNT]
    if not too_many_orders_df.empty:
        raise ValueError(
            f"Order count exceeds maximum of {MAX_ORDER_COUNT}: " f"{too_many_orders_df}"
        )

    return


@typechecked
def _validate_stop_no(df: pd.DataFrame) -> None:
    """Validate the stop number column."""
    _validate_col_not_empty(df=df, column=Columns.STOP_NO)
    _validate_greater_than_zero(df=df, column=Columns.STOP_NO)

    duplicates_df = df[df.duplicated(subset=[Columns.STOP_NO], keep=False)]
    if not duplicates_df.empty:
        raise ValueError(f"Duplicate stop numbers found: {duplicates_df}")

    stop_numbers = df[Columns.STOP_NO].to_list()
    if sorted(stop_numbers) != list(range(1, len(stop_numbers) + 1)):
        raise ValueError(f"Stop numbers are not contiguous starting at 1: {stop_numbers}")

    if stop_numbers != sorted(stop_numbers):
        raise ValueError(f"Stop numbers are not sorted: {stop_numbers}")

    return


@typechecked
def _validate_col_not_empty(df: pd.DataFrame, column: str) -> None:
    """No nulls or empty strings in column."""
    null_df = df[df[column].isnull()]
    if not null_df.empty:
        raise ValueError(f"Null values found in {column} column: " f"{null_df}")

    empty_df = df[df[column] == ""]
    if not empty_df.empty:
        raise ValueError(f"Empty values found in {column} column: " f"{empty_df}")

    return


@typechecked
def _validate_greater_than_zero(df: pd.DataFrame, column: str) -> None:
    """Validate column is greater than zero."""
    negative_df = df[df[column] <= 0]
    if not negative_df.empty:
        raise ValueError(
            f"Values less than or equal to zero found in {column} column: " f"{negative_df}"
        )

    return



"""Classes for making API calls."""

# TODO: Move this up to lib.

import logging
from abc import abstractmethod
from collections.abc import Callable
from time import sleep
from typing import Any

import requests
from requests.auth import HTTPBasicAuth
from typeguard import typechecked

from bfb_delivery.lib.constants import CIRCUIT_URL, CircuitColumns, RateLimits
from bfb_delivery.lib.dispatch.utils import get_circuit_key, get_response_dict

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


# TODO: Move base classes to comb_utils.


class BaseCaller:
    """An abstract class for making API calls.

    Example:
        .. code:: python

            class MyGetCaller(BaseCaller):
                target_response_value

                _min_wait_seconds: float = 0.1
                # Initialize _wait_seconds and timeout as a class variable.
                # Instances will adjust for class.
                _wait_seconds: float = _min_wait_seconds
                _timeout: float = 10

                def _set_request_call(self):
                    self._request_call = requests.get

                def _set_url(self):
                    self._url = "https://example.com/public/v0.2b/"

                def _handle_200(self):
                    super()._handle_200()
                    self.target_response_value = self.response_json["target_key"]

            my_caller = MyCaller()
            my_caller.call_api()
            target_response_value = my_caller.target_response_value

    .. important::
        You must initialize _wait_seconds, _timeout, and _min_wait_seconds in child classes.
        This allows child class instances to adjust the wait/timeout time for the child class.

    .. warning::
        There is a potential for this to run indefinitely for rate limiting and timeouts.
        It handles them somewhat intelligently, but the assumption is that someone is watching
        this run in the background and will stop it if it runs too long. It will eventually
        at least crash the memory, depending on available memory, mean time to failure, and
        time left in the universe.
    """

    # Set by object:
    #: The JSON from the response.
    response_json: dict[str, Any]
    #: The response from the API call.
    _response: requests.Response

    # Must set in child class with _set*:
    #: The requests call method. (get, post, etc.)
    _request_call: Callable
    #: The URL for the API call.
    _url: str

    # Must set in child class:
    #: The timeout for the API call.
    _timeout: float
    #: The minimum wait time between API calls.
    _min_wait_seconds: float
    #: The wait time between API calls. (Adjusted by instances, at class level.)
    _wait_seconds: float

    # Optionally set in child class, to pass to _request_call if needed:
    #: The kwargs to pass to the requests call.
    _call_kwargs: dict[str, Any] = {}
    #: The scalar to increase wait time on rate limiting.
    _wait_increase_scalar: float = RateLimits.WAIT_INCREASE_SCALAR
    #: The scalar to decrease wait time on success.
    _wait_decrease_scalar: float = RateLimits.WAIT_DECREASE_SECONDS

    @typechecked
    def __init__(self) -> None:  # noqa: ANN401
        """Initialize the BaseCaller object."""
        self._set_request_call()
        self._set_url()

    @abstractmethod
    @typechecked
    def _set_request_call(self) -> None:
        """Set the requests call method.

        requests.get, requests.post, etc.

        Raises:
            NotImplementedError: If not implemented in child class.
        """
        raise NotImplementedError

    @abstractmethod
    @typechecked
    def _set_url(self) -> None:
        """Set the URL for the API call.

        Raises:
            NotImplementedError: If not implemented in child class.
        """
        raise NotImplementedError

    @typechecked
    def call_api(self) -> None:
        """The main method for making the API call.

        Handle errors, parse response, and decrease class wait time on success.

        Raises:
            ValueError: If the response status code is not expected.
            requests.exceptions.HTTPError: For non-rate-limiting errors.
        """
        # Separated to allow for recursive calls on rate limiting.
        self._call_api()
        self._decrease_wait_time()

    @typechecked
    def _call_api(self) -> None:
        """Wait and make and handle the API call.

        Wrapped separately to allow for recursive calls on rate limiting and timeout.
        """
        sleep(type(self)._wait_seconds)
        self._make_call()
        self._raise_for_status()
        self._parse_response()

    @typechecked
    def _make_call(self) -> None:
        """Make the API call."""
        self._response = self._request_call(
            url=self._url,
            auth=HTTPBasicAuth(get_circuit_key(), ""),
            timeout=self._timeout,
            **self._call_kwargs,
        )

    @typechecked
    def _raise_for_status(self) -> None:
        """Handle error responses.

        For 429 (rate limiting), increases wait time and recursively calls the API.
        For timeout, increases timeout and recursively calls the API.

        Raises:
            requests.exceptions.HTTPError: For non-rate-limiting errors.
            requests.exceptions.Timeout: For timeouts.
        """
        try:
            self._response.raise_for_status()
        except requests.exceptions.HTTPError as http_e:
            if self._response.status_code == 429:
                self._handle_429()
            else:
                self._handle_unknown_error(e=http_e)
        except requests.exceptions.Timeout:
            self._handle_timeout()

    @typechecked
    def _parse_response(self) -> None:
        """Parse the non-error reponse (200).

        Raises:
            ValueError: If the response status code is not expected.
        """
        if self._response.status_code == 200:
            self._handle_200()
        elif self._response.status_code == 204:
            self._handle_204()
        elif self._response.status_code == 429:
            # This is here as well as in the _raise_for_status method because there was a case
            # when the status code was 429 but the response didn't raise.
            self._handle_429()
        else:
            response_dict = get_response_dict(response=self._response)
            raise ValueError(
                f"Unexpected response {self._response.status_code}:\n{response_dict}"
            )

    @typechecked
    def _handle_429(self) -> None:
        """Handle a 429 response.

        Inreases the class wait time and recursively calls the API.
        """
        self._increase_wait_time()
        logger.warning(f"Rate limited. Waiting {type(self)._wait_seconds} seconds to retry.")
        self._call_api()

    @typechecked
    def _handle_timeout(self) -> None:
        """Handle a timeout response.

        Increases the class timeout and recursively calls the API.
        """
        self._increase_timeout()
        response_dict = get_response_dict(response=self._response)
        logger.warning(
            f"Request timed out.\n{response_dict}"
            f"\nTrying again with longer timeout: {type(self)._timeout} seconds."
        )
        self._call_api()

    @typechecked
    def _handle_200(self) -> None:
        """Handle a 200 response.

        Just gets the JSON from the response and sets it to `response_json`.
        """
        self.response_json = self._response.json()

    @typechecked
    def _handle_204(self) -> None:
        """Handle a 204 response.

        Just sets `response_json` to an empty dictionary.
        """
        self.response_json = {}

    @typechecked
    def _handle_unknown_error(self, e: Exception) -> None:
        """Handle an unknown error response.

        Raises:
            Exception: The original error.
        """
        response_dict = get_response_dict(response=self._response)
        err_msg = f"Got {self._response.status_code} response:\n{response_dict}"
        raise requests.exceptions.HTTPError(err_msg) from e

    @typechecked
    def _decrease_wait_time(self) -> None:
        """Decrease the wait time between API calls for whole class."""
        cls = type(self)
        cls._wait_seconds = max(
            cls._wait_seconds * self._wait_decrease_scalar, cls._min_wait_seconds
        )

    @typechecked
    def _increase_wait_time(self) -> None:
        """Increase the wait time between API calls for whole class."""
        cls = type(self)
        cls._wait_seconds = cls._wait_seconds * self._wait_increase_scalar

    @typechecked
    def _increase_timeout(self) -> None:
        """Increase the timeout for the API call for whole class."""
        cls = type(self)
        cls._timeout = cls._timeout * self._wait_increase_scalar


class BaseGetCaller(BaseCaller):
    """A base class for making GET API calls.

    Presets the timeout, initial wait time, and requests method.
    """

    _timeout: float = RateLimits.READ_TIMEOUT_SECONDS
    _min_wait_seconds: float = RateLimits.READ_SECONDS
    _wait_seconds: float = _min_wait_seconds

    @typechecked
    def _set_request_call(self) -> None:
        """Set the requests call method to `requests.get`."""
        self._request_call = requests.get


class BasePostCaller(BaseCaller):
    """A base class for making POST API calls.

    Presets the timeout, initial wait time, and requests method.
    """

    _timeout: float = RateLimits.WRITE_TIMEOUT_SECONDS
    _min_wait_seconds: float = RateLimits.WRITE_SECONDS
    _wait_seconds: float = _min_wait_seconds

    @typechecked
    def _set_request_call(self) -> None:
        """Set the requests call method to `requests.post`."""
        self._request_call = requests.post


class BaseDeleteCaller(BasePostCaller):
    """A base class for making DELETE API calls.

    Presets the timeout, initial wait time, and requests method.
    """

    @typechecked
    def _set_request_call(self) -> None:
        """Set the requests call method to `requests.delete`."""
        self._request_call = requests.delete


class BaseOptimizationCaller(BaseCaller):
    """A base class for checking the status of an optimization."""

    #: The ID of the operation.
    operation_id: str
    #: Whether the optimization is finished.
    finished: bool

    _timeout: float = RateLimits.WRITE_TIMEOUT_SECONDS
    _min_wait_seconds: float = RateLimits.OPTIMIZATION_PER_SECOND
    _wait_seconds: float = _min_wait_seconds

    #: The ID of the plan.
    _plan_id: str
    #: The title of the plan.
    _plan_title: str

    @typechecked
    def __init__(self, plan_id: str, plan_title: str) -> None:  # noqa: ANN401
        """Initialize the BaseOptimizationCaller object.

        Args:
            plan_id: The ID of the plan. (e.g. plans/asfoghaev)
            plan_title: The title of the plan.
        """
        self._plan_id = plan_id
        self._plan_title = plan_title
        super().__init__()

    @typechecked
    def _handle_200(self) -> None:
        """Handle a 200 response.

        Sets `operation_id` and whether the optimization is `finished`.

        Raises:
            RuntimeError: If the optimization was canceled, stops were skipped, or there were
                errors.
        """
        super()._handle_200()

        if self.response_json[CircuitColumns.METADATA][CircuitColumns.CANCELED]:
            raise RuntimeError(
                f"Optimization canceled for {self._plan_title} ({self._plan_id}):"
                f"\n{self.response_json}"
            )
        if self.response_json.get(CircuitColumns.RESULT):
            if self.response_json[CircuitColumns.RESULT].get(CircuitColumns.SKIPPED_STOPS):
                raise RuntimeError(
                    f"Skipped optimization stops for {self._plan_title} ({self._plan_id}):"
                    f"\n{self.response_json}"
                )
            if self.response_json[CircuitColumns.RESULT].get(CircuitColumns.CODE):
                raise RuntimeError(
                    f"Errors in optimization for {self._plan_title} ({self._plan_id}):"
                    f"\n{self.response_json}"
                )

        self.operation_id = self.response_json[CircuitColumns.ID]
        self.finished = self.response_json[CircuitColumns.DONE]


class PagedResponseGetter(BaseGetCaller):
    """Class for getting paged responses."""

    # The nextPageToken returned, but called salsa to avoid bandit.
    next_page_salsa: str | None

    #: The URL for the page.
    _page_url: str

    @typechecked
    def __init__(self, page_url: str) -> None:
        """Initialize the PagedResponseGetter object.

        Args:
            page_url: The URL for the page. (Optionally contains nextPageToken.)
        """
        self._page_url = page_url
        super().__init__()

    @typechecked
    def _set_url(self) -> None:
        """Set the URL for the API call to the `page_url`."""
        self._url = self._page_url

    @typechecked
    def _handle_200(self) -> None:
        """Handle a 200 response.

        Sets `next_page_salsa` to the nextPageToken.
        """
        super()._handle_200()
        self.next_page_salsa = self.response_json.get("nextPageToken", None)


class PlanInitializer(BasePostCaller):
    """Class for initializing plans."""

    #: The ID of the plan.
    plan_id: str
    #: Whether the plan is writeable.
    writable: bool

    #: The data dictionary for the plan.
    _plan_data: dict

    @typechecked
    def __init__(self, plan_data: dict) -> None:
        """Initialize the PlanInitializer object.

        Args:
            plan_data: The data dictionary for the plan.
                To pass to `requests.post` `json` param.
        """
        self._plan_data = plan_data
        self._call_kwargs = {"json": plan_data}
        super().__init__()

    @typechecked
    def _set_url(self) -> None:
        """Set the URL for the API call."""
        self._url = f"{CIRCUIT_URL}/plans"

    @typechecked
    def _handle_200(self) -> None:
        """Handle a 200 response.

        Sets `plan_id` and `writable`.
        """
        super()._handle_200()
        self.plan_id = self.response_json[CircuitColumns.ID]
        self.writable = self.response_json[CircuitColumns.WRITABLE]


class StopUploader(BasePostCaller):
    """Class for batch uploading stops."""

    stop_ids: list[str]

    _min_wait_seconds: float = RateLimits.BATCH_STOP_IMPORT_SECONDS
    _wait_seconds: float = _min_wait_seconds

    _plan_id: str
    _plan_title: str

    @typechecked
    def __init__(
        self,
        plan_id: str,
        plan_title: str,
        stop_array: list[dict[str, dict[str, str] | list[str] | int | str]],
    ) -> None:
        """Initialize the StopUploader object.

        Args:
            plan_id: The ID of the plan. (e.g. plans/asfoghaev)
            plan_title: The title of the plan.
            stop_array: The array of stops dictionaries to upload.
                To pass to `requests.post` `json` param.
        """
        self._plan_id = plan_id
        self._plan_title = plan_title
        self._stop_array = stop_array
        self._call_kwargs = {"json": stop_array}
        super().__init__()

    @typechecked
    def _set_url(self) -> None:
        """Set the URL for the API call with `plan_id`."""
        self._url = f"{CIRCUIT_URL}/{self._plan_id}/stops:import"

    @typechecked
    def _handle_200(self) -> None:
        """Handle a 200 response.

        Sets `stop_ids` to the successful stop IDs.

        Raises:
            RuntimeError: If stops failed to upload.
            RuntimeError: If the number of stops uploaded differs from input.
        """
        super()._handle_200()

        self.stop_ids = self.response_json["success"]
        failed = self.response_json.get("failed")
        if failed:
            raise RuntimeError(
                f"For {self._plan_title} ({self._plan_id}), failed to upload stops:\n{failed}"
            )
        elif len(self.stop_ids) != len(self._stop_array):
            raise RuntimeError(
                f"For {self._plan_title} ({self._plan_id}), did not upload same number of "
                f"stops as input:\n{self.stop_ids}\n{self._stop_array}"
            )


class OptimizationLauncher(BaseOptimizationCaller, BasePostCaller):
    """A class for launching route optimization.

    Args:
        plan_id: The ID of the plan. (e.g. plans/asfoghaev)
        plan_title: The title of the plan.
    """

    @typechecked
    def _set_url(self) -> None:
        """Set the URL for the API call with the `plan_id`."""
        self._url = f"{CIRCUIT_URL}/{self._plan_id}:optimize"


class OptimizationChecker(BaseOptimizationCaller, BaseGetCaller):
    """A class for checking the status of an optimization."""

    _timeout: float = RateLimits.READ_TIMEOUT_SECONDS
    _min_wait_seconds: float = RateLimits.READ_SECONDS
    _wait_seconds: float = _min_wait_seconds

    @typechecked
    def __init__(self, plan_id: str, plan_title: str, operation_id: str) -> None:
        """Initialize the OptimizationChecker object.

        Args:
            plan_id: The ID of the plan.
            plan_title: The title of the plan.
            operation_id: The ID of the operation.
        """
        self.operation_id = operation_id
        super().__init__(plan_id=plan_id, plan_title=plan_title)

    @typechecked
    def _set_url(self) -> None:
        """Set the URL for the API call."""
        self._url = f"{CIRCUIT_URL}/{self.operation_id}"


class PlanDistributor(BasePostCaller):
    """Class for distributing plans."""

    distributed: bool

    #: The ID of the plan.
    _plan_id: str
    #: The title of the plan
    _plan_title: str

    @typechecked
    def __init__(self, plan_id: str, plan_title: str) -> None:
        """Initialize the PlanDistributor object.

        Args:
            plan_id: The ID of the plan. (e.g. plans/asfoghaev)
            plan_title: The title of the plan.
        """
        self._plan_id = plan_id
        self._plan_title = plan_title
        super().__init__()

    @typechecked
    def _set_url(self) -> None:
        """Set the URL for the API call with the `plan_id`."""
        self._url = f"{CIRCUIT_URL}/{self._plan_id}:distribute"

    @typechecked
    def _handle_200(self) -> None:
        """Handle a 200 response.

        Raises:
            RuntimeError: If the plan was not distributed.
        """
        super()._handle_200()
        self.distributed = self.response_json[CircuitColumns.DISTRIBUTED]
        if not self.distributed:
            raise RuntimeError(
                f"Failed to distribute plan {self._plan_title} ({self._plan_id}):"
                f"\n{self.response_json}"
            )


class PlanDeleter(BaseDeleteCaller):
    """Class for deleting plans."""

    #: Whether the plan was deleted.
    deletion: bool = False

    @typechecked
    def __init__(self, plan_id: str) -> None:
        """Initialize the PlanDeleter object.

        Args:
            plan_id: The ID of the plan.
        """
        self._plan_id = plan_id
        super().__init__()

    @typechecked
    def _set_url(self) -> None:
        """Set the URL for the API call with the `plan_id`."""
        self._url = f"{CIRCUIT_URL}/{self._plan_id}"

    @typechecked
    def _handle_204(self) -> None:
        """Handle a 204 response.

        Sets `deletion` to True.
        """
        super()._handle_204()
        self.deletion = True



"""Internal functions overlay library and are typically wrapped by public functions.

This allows us to maintain a separation of API from implementation.
Internal functions may come with extra options that public functions don't have, say for
power users and developers who may want to use an existing DB session or something.
"""

from pathlib import Path

from typeguard import typechecked

from bfb_delivery.lib.constants import DocStrings
from bfb_delivery.lib.dispatch import write_to_circuit
from bfb_delivery.lib.dispatch.read_circuit import get_route_files
from bfb_delivery.lib.formatting import sheet_shaping


def build_routes_from_chunked(  # noqa: D103
    input_path: str,
    output_dir: str,
    start_date: str,
    no_distribute: bool,
    verbose: bool,
    extra_notes_file: str,
) -> Path:
    return write_to_circuit.build_routes_from_chunked(
        input_path=input_path,
        output_dir=output_dir,
        start_date=start_date,
        no_distribute=no_distribute,
        verbose=verbose,
        extra_notes_file=extra_notes_file,
    )


build_routes_from_chunked.__doc__ = DocStrings.BUILD_ROUTES_FROM_CHUNKED.api_docstring


@typechecked
def split_chunked_route(  # noqa: D103
    input_path: Path | str,
    output_dir: Path | str,
    output_filename: str,
    n_books: int,
    book_one_drivers_file: str,
    date: str,
) -> list[Path]:
    return sheet_shaping.split_chunked_route(
        input_path=input_path,
        output_dir=output_dir,
        output_filename=output_filename,
        n_books=n_books,
        book_one_drivers_file=book_one_drivers_file,
        date=date,
    )


split_chunked_route.__doc__ = DocStrings.SPLIT_CHUNKED_ROUTE.api_docstring


@typechecked
def create_manifests_from_circuit(  # noqa: D103
    start_date: str,
    end_date: str,
    plan_ids: list[str],
    output_dir: str,
    output_filename: str,
    circuit_output_dir: str,
    all_hhs: bool,
    verbose: bool,
    extra_notes_file: str,
) -> tuple[Path, Path]:
    circuit_output_dir = get_route_files(
        start_date=start_date,
        end_date=end_date,
        plan_ids=plan_ids,
        output_dir=circuit_output_dir,
        all_hhs=all_hhs,
        verbose=verbose,
    )
    formatted_manifest_path = sheet_shaping.create_manifests(
        input_dir=circuit_output_dir,
        output_dir=output_dir,
        output_filename=output_filename,
        extra_notes_file=extra_notes_file,
    )

    return formatted_manifest_path, Path(circuit_output_dir)


create_manifests_from_circuit.__doc__ = DocStrings.CREATE_MANIFESTS_FROM_CIRCUIT.api_docstring


@typechecked
def create_manifests(  # noqa: D103
    input_dir: Path | str, output_dir: Path | str, output_filename: str, extra_notes_file: str
) -> Path:
    formatted_manifest_path = sheet_shaping.create_manifests(
        input_dir=input_dir,
        output_dir=output_dir,
        output_filename=output_filename,
        extra_notes_file=extra_notes_file,
    )

    return formatted_manifest_path


create_manifests.__doc__ = DocStrings.CREATE_MANIFESTS.api_docstring


@typechecked
def combine_route_tables(  # noqa: D103
    input_dir: Path | str, output_dir: Path | str, output_filename: str
) -> Path:
    return sheet_shaping.combine_route_tables(
        input_dir=input_dir, output_dir=output_dir, output_filename=output_filename
    )


combine_route_tables.__doc__ = DocStrings.COMBINE_ROUTE_TABLES.api_docstring


@typechecked
def format_combined_routes(  # noqa: D103
    input_path: Path | str,
    output_dir: Path | str,
    output_filename: str,
    extra_notes_file: str,
) -> Path:
    return sheet_shaping.format_combined_routes(
        input_path=input_path,
        output_dir=output_dir,
        output_filename=output_filename,
        extra_notes_file=extra_notes_file,
    )


format_combined_routes.__doc__ = DocStrings.FORMAT_COMBINED_ROUTES.api_docstring



"""Functions for shaping and formatting spreadsheets."""

import logging
import math
import warnings
from datetime import datetime
from pathlib import Path
from typing import Any

import pandas as pd
from openpyxl import Workbook
from openpyxl.styles import Alignment, Border, Font, PatternFill, Side
from openpyxl.utils.dataframe import dataframe_to_rows
from openpyxl.worksheet.worksheet import Worksheet
from typeguard import typechecked

from bfb_delivery.lib.constants import (
    ADDRESS_COLUMN_WIDTH,
    BOX_TYPE_COLOR_MAP,
    COLUMN_NAME_MAP,
    COMBINED_ROUTES_COLUMNS,
    FILE_DATE_FORMAT,
    FORMATTED_ROUTES_COLUMNS,
    MANIFEST_DATE_FORMAT,
    NOTES_COLUMN_WIDTH,
    PROTEIN_BOX_TYPES,
    SPLIT_ROUTE_COLUMNS,
    BoxType,
    CellColors,
    Columns,
    DocStrings,
)
from bfb_delivery.lib.formatting.data_cleaning import (
    format_and_validate_data,
    format_column_names,
)
from bfb_delivery.lib.formatting.utils import (
    get_book_one_drivers,
    get_extra_notes,
    get_phone_number,
    map_columns,
)
from bfb_delivery.lib.utils import get_friday

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)

# Silences warning for in-place operations on copied df slices.
pd.options.mode.copy_on_write = True


# TODO: Use Pandera.
# TODO: Switch to or allow CSVs instead of Excel files.
@typechecked
def split_chunked_route(  # noqa: D103
    input_path: Path | str,
    output_dir: Path | str,
    output_filename: str,
    n_books: int,
    book_one_drivers_file: str,
    date: str,
) -> list[Path]:
    if n_books <= 0:
        raise ValueError("n_books must be greater than 0.")
    # TODO: Make this accept input_path only as Path? Or only as str to simplify?
    input_path = Path(input_path)
    date = date if date else get_friday(fmt=MANIFEST_DATE_FORMAT)

    chunked_sheet: pd.DataFrame = pd.read_excel(input_path)
    chunked_sheet.columns = format_column_names(columns=chunked_sheet.columns.to_list())
    map_columns(df=chunked_sheet, column_name_map=COLUMN_NAME_MAP, invert_map=False)
    format_and_validate_data(df=chunked_sheet, columns=SPLIT_ROUTE_COLUMNS + [Columns.DRIVER])
    chunked_sheet.sort_values(by=[Columns.DRIVER, Columns.STOP_NO], inplace=True)
    # TODO: Validate columns? (Use Pandera?)

    drivers = list(chunked_sheet[Columns.DRIVER].unique())
    driver_count = len(drivers)
    if driver_count < n_books:
        raise ValueError(
            "n_books must be less than or equal to the number of drivers: "
            f"driver_count: {driver_count}, n_books: {n_books}."
        )

    output_dir = Path(output_dir) if output_dir else Path(input_path).parent
    base_output_filename = (
        f"split_workbook_{datetime.now().strftime(FILE_DATE_FORMAT)}.xlsx"
        if output_filename == ""
        else output_filename
    )
    output_dir.mkdir(parents=True, exist_ok=True)

    split_workbook_paths: list[Path] = []
    driver_sets = _get_driver_sets(
        drivers=drivers, n_books=n_books, book_one_drivers_file=book_one_drivers_file
    )

    logger.info(f"Writing split chunked workbooks to {output_dir.resolve()}")
    for i, driver_set in enumerate(driver_sets):
        i_file_name = f"{base_output_filename.split('.')[0]}_{i + 1}.xlsx"
        split_workbook_path: Path = output_dir / i_file_name
        split_workbook_paths.append(split_workbook_path)

        driver_set_df = chunked_sheet[chunked_sheet[Columns.DRIVER].isin(driver_set)]
        driver_set_df.sort_values(by=[Columns.DRIVER, Columns.STOP_NO], inplace=True)

        with pd.ExcelWriter(split_workbook_path) as writer:
            for driver_name, data in driver_set_df.groupby(Columns.DRIVER):
                data[SPLIT_ROUTE_COLUMNS].to_excel(
                    writer, sheet_name=f"{date} {driver_name}", index=False
                )

    split_workbook_paths = [path.resolve() for path in split_workbook_paths]

    return split_workbook_paths


split_chunked_route.__doc__ = DocStrings.SPLIT_CHUNKED_ROUTE.api_docstring


@typechecked
def create_manifests(  # noqa: D103
    input_dir: Path | str, output_dir: Path | str, output_filename: str, extra_notes_file: str
) -> Path:
    output_filename = (
        f"final_manifests_{datetime.now().strftime(FILE_DATE_FORMAT)}.xlsx"
        if output_filename == ""
        else output_filename
    )

    combined_route_workbook_path = combine_route_tables(
        input_dir=input_dir, output_dir=output_dir, output_filename=""
    )

    formatted_manifest_path = format_combined_routes(
        input_path=combined_route_workbook_path,
        output_dir=output_dir,
        output_filename=output_filename,
        extra_notes_file=extra_notes_file,
    )

    return formatted_manifest_path


create_manifests.__doc__ = DocStrings.CREATE_MANIFESTS.api_docstring


@typechecked
def combine_route_tables(  # noqa: D103
    input_dir: Path | str, output_dir: Path | str, output_filename: str
) -> Path:
    input_dir = Path(input_dir)
    paths = list(input_dir.glob("*.csv"))

    output_dir = Path(output_dir) if output_dir else paths[0].parent
    output_filename = (
        f"combined_routes_{datetime.now().strftime(FILE_DATE_FORMAT)}.xlsx"
        if output_filename == ""
        else output_filename
    )
    output_path = output_dir / output_filename
    output_dir.mkdir(parents=True, exist_ok=True)

    logger.info(f"Writing combined routes to {output_path.resolve()}")
    with pd.ExcelWriter(output_path) as writer:
        for path in sorted(paths):
            route_df = pd.read_csv(path)
            map_columns(df=route_df, column_name_map=COLUMN_NAME_MAP, invert_map=True)
            route_df.sort_values(by=[Columns.STOP_NO], inplace=True)
            route_df[COMBINED_ROUTES_COLUMNS].to_excel(
                writer, sheet_name=path.stem, index=False
            )

    return output_path.resolve()


combine_route_tables.__doc__ = DocStrings.COMBINE_ROUTE_TABLES.api_docstring


@typechecked
def format_combined_routes(  # noqa: D103
    input_path: Path | str,
    output_dir: Path | str,
    output_filename: str,
    extra_notes_file: str,
) -> Path:
    input_path = Path(input_path)
    output_dir = Path(output_dir) if output_dir else input_path.parent
    output_filename = (
        f"formatted_routes_{datetime.now().strftime(FILE_DATE_FORMAT)}.xlsx"
        if output_filename == ""
        else output_filename
    )
    output_path = Path(output_dir) / output_filename

    output_dir.mkdir(parents=True, exist_ok=True)

    extra_notes_df = get_extra_notes(file_path=extra_notes_file)

    wb = Workbook()
    wb.remove(wb["Sheet"])
    with pd.ExcelFile(input_path) as xls:
        for sheet_idx, sheet_name in enumerate(sorted(xls.sheet_names)):

            route_df = pd.read_excel(xls, sheet_name)

            # TODO: Use Pandera?
            route_df.columns = format_column_names(columns=route_df.columns.to_list())
            format_and_validate_data(df=route_df, columns=COMBINED_ROUTES_COLUMNS)

            # TODO: Order by apartment number, and redo stop numbers?
            # May need to postpone this.
            # Or, for now, just do it if the apartments are already in contiguous stops.
            # Or if discontinuous, just regroup and bump the following stops.
            # Also, may not make the most sense in order of apt number. Ask team.
            route_df.sort_values(by=[Columns.STOP_NO], inplace=True)

            agg_dict = _aggregate_route_data(df=route_df, extra_notes_df=extra_notes_df)
            # TODO: !! What happens when there are more than one order for a stop? Two rows?
            # (Since order count column is dropped in manifest)
            # Oh wait, they're all 1s, so is that just a way for them to count them with sum?
            # If that's so, ignore it or validate always a 1?

            _make_manifest_sheet(
                wb=wb,
                agg_dict=agg_dict,
                route_df=route_df,
                sheet_name=str(sheet_name),
                sheet_idx=sheet_idx,
            )

    # Can check cell values, though. (Maye read dataframe from start row?)
    logger.info(f"Writing formatted routes to {output_path.resolve()}")
    wb.save(output_path)

    return output_path.resolve()


format_combined_routes.__doc__ = DocStrings.FORMAT_COMBINED_ROUTES.api_docstring


@typechecked
def _get_driver_sets(
    drivers: list[str], n_books: int, book_one_drivers_file: str
) -> list[list[str]]:
    """Split drivers into n_books sets."""
    drivers = sorted(drivers)
    drivers = _move_book_one_drivers_to_front(
        drivers=drivers, book_one_drivers_file=book_one_drivers_file
    )
    driver_sets = _split_driver_list(drivers=drivers, n_books=n_books)
    driver_sets = _group_numbered_drivers(driver_sets=driver_sets)
    driver_sets = [sorted(driver_set) for driver_set in driver_sets]

    return driver_sets


@typechecked
def _move_book_one_drivers_to_front(
    drivers: list[str], book_one_drivers_file: str
) -> list[str]:
    """Move book one drivers to the front of the list."""
    book_one_drivers = get_book_one_drivers(file_path=book_one_drivers_file)
    drivers = [d for d in book_one_drivers if d in drivers] + [
        d for d in drivers if d not in book_one_drivers
    ]

    return drivers


@typechecked
def _split_driver_list(drivers: list[str], n_books: int) -> list[list[str]]:
    """Split drivers into n_books sets, in order passed."""
    driver_sets = []

    len_drivers = len(drivers)
    start_idx = 0
    inc = len_drivers // n_books
    remainder = len_drivers % n_books

    while start_idx < len_drivers:
        end_idx = start_idx + inc
        if remainder > 0:
            end_idx += 1
            remainder -= 1
        end_idx = end_idx if end_idx < len_drivers else len_drivers

        driver_sets.append(drivers[start_idx:end_idx])
        start_idx = end_idx

    return driver_sets


@typechecked
def _group_numbered_drivers(driver_sets: list[list[str]]) -> list[list[str]]:
    """Merge drivers with numbers into a single set."""
    driver_sets_updated = driver_sets.copy()

    updated = True
    while updated:
        driver_sets_copy = driver_sets_updated.copy()
        for i, driver_set in enumerate(driver_sets_updated):
            driver_set_i = driver_set.copy()
            numbered_drivers = [d for d in driver_set if "#" in d]

            for d in numbered_drivers:
                driver_name = d.split("#")[0].strip()

                i_plus_1 = i + 1
                if i_plus_1 < len(driver_sets_updated):
                    for j, driver_set_j in enumerate(
                        driver_sets_updated[i_plus_1:], start=i_plus_1
                    ):
                        matching_drivers = [d for d in driver_set_j if driver_name in d]
                        driver_set_i += matching_drivers
                        driver_sets_updated[j] = [
                            d for d in driver_set_j if d not in matching_drivers
                        ]
            driver_sets_updated[i] = driver_set_i

        updated = driver_sets_copy != driver_sets_updated

    driver_sets_updated = [driver_set for driver_set in driver_sets_updated if driver_set]

    return driver_sets_updated


@typechecked
def _aggregate_route_data(df: pd.DataFrame, extra_notes_df: pd.DataFrame) -> dict[str, Any]:
    """Aggregate data for a single route.

    Args:
        df: The route data to aggregate.
        extra_notes_df: Extra notes to include in the manifest if tagged.

    Returns:
        Dictionary of aggregated data.
    """
    df = df.copy()

    df[Columns.BOX_TYPE] = df[Columns.BOX_TYPE].str.upper().str.strip()
    box_types = df[Columns.BOX_TYPE].unique()
    extra_box_types = set(box_types) - set(BoxType)
    if extra_box_types:
        raise ValueError(f"Invalid box type in route data: {extra_box_types}")

    extra_notes_list = []
    used_tags = []
    for _, row in extra_notes_df.iterrows():
        for notes in df[Columns.NOTES]:
            if row["tag"].upper() in notes.upper() and row["tag"] not in used_tags:
                used_tags.append(row["tag"])
                note = "* " + row["tag"].replace("*", "").strip() + ": " + row["note"]
                extra_notes_list.append(note)

    agg_dict = {
        "box_counts": df.groupby(Columns.BOX_TYPE)[Columns.ORDER_COUNT].sum().to_dict(),
        "total_box_count": df[Columns.ORDER_COUNT].sum(),
        "protein_box_count": df[df[Columns.BOX_TYPE].isin(PROTEIN_BOX_TYPES)][
            Columns.ORDER_COUNT
        ].sum(),
        "neighborhoods": df[Columns.NEIGHBORHOOD].unique().tolist(),
        "extra_notes": extra_notes_list,
    }

    for box_type in BoxType:
        if box_type.value not in agg_dict["box_counts"]:
            agg_dict["box_counts"][box_type.value] = 0

    return agg_dict


@typechecked
def _make_manifest_sheet(
    wb: Workbook, agg_dict: dict, route_df: pd.DataFrame, sheet_name: str, sheet_idx: int
) -> None:
    """Create a manifest sheet."""
    ws = wb.create_sheet(title=str(sheet_name), index=sheet_idx)
    _add_header_row(ws=ws)
    neighborhoods_row_number = _add_aggregate_block(
        ws=ws, agg_dict=agg_dict, sheet_name=sheet_name
    )
    df_start_row = _write_data_to_sheet(ws=ws, df=route_df)
    _auto_adjust_column_widths(ws=ws, df_start_row=df_start_row)
    _word_wrap_columns(ws=ws)
    _merge_and_wrap_neighborhoods(ws=ws, neighborhoods_row_number=neighborhoods_row_number)
    _append_extra_notes(ws=ws, extra_notes=agg_dict["extra_notes"])

    # TODO: Set print_area (Use calculate_dimensions)
    # TODO: set_printer_settings(paper_size, orientation)


@typechecked
def _add_header_row(ws: Worksheet) -> None:
    """Append a reusable formatted row to the worksheet."""
    font = Font(bold=True)
    alignment_left = Alignment(horizontal="left")
    alignment_right = Alignment(horizontal="right")
    fill = PatternFill(
        start_color=CellColors.HEADER, end_color=CellColors.HEADER, fill_type="solid"
    )

    driver_support_phone = get_phone_number("driver_support")
    recipient_support_phone = get_phone_number("recipient_support")
    formatted_row = [
        {
            "value": f"DRIVER SUPPORT: {driver_support_phone}",
            "font": font,
            "alignment": alignment_left,
            "fill": fill,
        },
        {"value": "", "font": font, "alignment": None, "fill": fill},
        {"value": "", "font": font, "alignment": None, "fill": fill},
        {
            "value": f"RECIPIENT SUPPORT: {recipient_support_phone}",
            "font": font,
            "alignment": alignment_right,
            "fill": fill,
        },
        {"value": "", "font": font, "alignment": None, "fill": fill},
        {
            "value": "PLEASE SHRED MANIFEST AFTER COMPLETING ROUTE.",
            "font": font,
            "alignment": alignment_right,
            "fill": fill,
        },
    ]

    for col_idx, col_data in enumerate(formatted_row, start=1):
        cell_value = col_data["value"] if isinstance(col_data["value"], str) else ""
        cell = ws.cell(row=1, column=col_idx, value=cell_value)
        if col_data["font"]:
            if isinstance(col_data["font"], Font):
                cell.font = col_data["font"]
        if col_data["alignment"]:
            if isinstance(col_data["alignment"], Alignment):
                cell.alignment = col_data["alignment"]
        if col_data["fill"]:
            if isinstance(col_data["fill"], PatternFill):
                cell.fill = col_data["fill"]

    return


@typechecked
def _add_aggregate_block(ws: Worksheet, agg_dict: dict, sheet_name: str) -> int:
    """Append left and right aggregation blocks to the worksheet row by row."""
    date = str(sheet_name).split(" ")[0]
    driver_name = " ".join(str(sheet_name).split(" ")[1:])

    # TODO: Yeah, let's use an enum for box types since the manifest is a contract.
    thin_border = Border(
        left=Side(style="thin"),
        right=Side(style="thin"),
        top=Side(style="thin"),
        bottom=Side(style="thin"),
    )
    alignment_left = Alignment(horizontal="left")
    alignment_right = Alignment(horizontal="right")
    bold_font = Font(bold=True)

    left_block = _get_left_block(date=date, driver_name=driver_name, agg_dict=agg_dict)
    right_block = _get_right_block(thin_border=thin_border, agg_dict=agg_dict)

    start_row = ws.max_row + 1
    neighborhoods_row_number = 0
    for i, (left_row, right_row) in enumerate(
        zip(left_block, right_block, strict=True), start=start_row
    ):
        for col_idx, cell_definition in enumerate(left_row, start=1):
            cell = ws.cell(row=i, column=col_idx, value=cell_definition["value"])
            cell.font = bold_font
            cell.alignment = alignment_left
            if cell_definition["value"] and cell_definition["value"].startswith(
                "Neighborhoods"
            ):
                neighborhoods_row_number = i

        for col_idx, cell_definition in enumerate(right_row, start=5):
            cell = ws.cell(row=i, column=col_idx, value=cell_definition["value"])
            cell.font = bold_font
            cell.alignment = alignment_right
            if isinstance(cell_definition["fill"], PatternFill):
                cell.fill = cell_definition["fill"]
            if isinstance(cell_definition["border"], Border):
                cell.border = cell_definition["border"]

    return neighborhoods_row_number


def _get_left_block(
    date: str, driver_name: str, agg_dict: dict
) -> list[list[dict[str, None]] | list[dict[str, str]]]:
    left_block = [
        [{"value": None}],
        [{"value": f"Date: {date}"}],
        [{"value": None}],
        [{"value": f"Driver: {driver_name}"}],
        [{"value": None}],
        [{"value": f"Neighborhoods: {', '.join(agg_dict['neighborhoods'])}"}],
        [{"value": None}],
    ]

    return left_block


def _get_right_block(thin_border: Border, agg_dict: dict) -> list[list[dict]]:
    right_block = [
        [{"value": None, "fill": None, "border": None}],
        [
            {
                "value": "BASIC",
                "fill": PatternFill(
                    start_color=CellColors.BASIC,
                    end_color=CellColors.BASIC,
                    fill_type="solid",
                ),
                "border": thin_border,
            },
            {
                "value": agg_dict["box_counts"].get("BASIC", 0),
                "fill": None,
                "border": thin_border,
            },
        ],
        [
            {
                "value": "LA",
                "fill": PatternFill(
                    start_color=CellColors.LA, end_color=CellColors.LA, fill_type="solid"
                ),
                "border": thin_border,
            },
            {
                "value": agg_dict["box_counts"].get("LA", 0),
                "fill": None,
                "border": thin_border,
            },
        ],
        [
            {
                "value": "GF",
                "fill": PatternFill(
                    start_color=CellColors.GF, end_color=CellColors.GF, fill_type="solid"
                ),
                "border": thin_border,
            },
            {
                "value": agg_dict["box_counts"].get("GF", 0),
                "fill": None,
                "border": thin_border,
            },
        ],
        [
            {
                "value": "VEGAN",
                "fill": PatternFill(
                    start_color=CellColors.VEGAN,
                    end_color=CellColors.VEGAN,
                    fill_type="solid",
                ),
                "border": thin_border,
            },
            {
                "value": agg_dict["box_counts"].get("VEGAN", 0),
                "fill": None,
                "border": thin_border,
            },
        ],
        [
            {"value": "TOTAL BOX COUNT=", "fill": None, "border": None},
            {"value": agg_dict["total_box_count"], "fill": None, "border": None},
        ],
        [
            {"value": "PROTEIN COUNT=", "fill": None, "border": None},
            {"value": agg_dict["protein_box_count"], "fill": None, "border": None},
        ],
    ]

    return right_block


@typechecked
def _write_data_to_sheet(ws: Worksheet, df: pd.DataFrame) -> int:
    """Write and format the dataframe itself."""
    thin_border = Border(
        left=Side(style="thin"),
        right=Side(style="thin"),
        top=Side(style="thin"),
        bottom=Side(style="thin"),
    )

    header_font = Font(bold=True)

    box_type_col_idx = df.columns.get_loc(Columns.BOX_TYPE)

    start_row = ws.max_row + 1
    for r_idx, row in enumerate(
        dataframe_to_rows(df[FORMATTED_ROUTES_COLUMNS], index=False, header=True),
        start=start_row,
    ):
        for c_idx, value in enumerate(row, start=1):
            cell = ws.cell(row=r_idx, column=c_idx, value=value)
            cell.border = thin_border
            if r_idx == start_row:
                cell.font = header_font
                cell.alignment = Alignment(horizontal="left")

            if c_idx == box_type_col_idx and r_idx > start_row:
                box_type = str(value)
                fill_color = BOX_TYPE_COLOR_MAP.get(box_type)
                if fill_color:
                    cell.fill = PatternFill(
                        start_color=fill_color, end_color=fill_color, fill_type="solid"
                    )

    return start_row


@typechecked
def _auto_adjust_column_widths(ws: Worksheet, df_start_row: int) -> None:
    """Auto-adjust column widths to fit the dataframe."""
    for col in ws.columns:
        max_length = 0

        col_letter = col[0].column_letter
        padding_scalar = 1
        for cell in col:
            if cell.row >= df_start_row:
                try:
                    if cell.value:
                        max_length = max(max_length, len(str(cell.value)) * padding_scalar)
                except Exception as e:
                    warnings.warn(f"Error while adjusting column widths: {e}", stacklevel=2)

        ws.column_dimensions[col_letter].width = max(8, round(max_length))

    return


@typechecked
def _word_wrap_columns(ws: Worksheet) -> None:
    """Word wrap the notes column, and set width."""
    start_row = 10
    end_row = ws.max_row
    _word_wrap_column(
        ws=ws,
        start_row=start_row,
        end_row=end_row,
        col_letter="C",
        width=ADDRESS_COLUMN_WIDTH,
    )
    _word_wrap_column(
        ws=ws, start_row=start_row, end_row=end_row, col_letter="E", width=NOTES_COLUMN_WIDTH
    )

    return


@typechecked
def _word_wrap_column(
    ws: Worksheet, start_row: int, end_row: int, col_letter: str, width: float
) -> None:
    """Word wrap column, and set width."""
    ws.column_dimensions[col_letter].width = width
    for row in ws[f"{col_letter}{start_row}:{col_letter}{end_row}"]:
        for cell in row:
            cell.alignment = Alignment(wrap_text=True)

    return


@typechecked
def _merge_and_wrap_neighborhoods(ws: Worksheet, neighborhoods_row_number: int) -> None:
    """Merge the neighborhoods cell and wrap the text."""
    start_col = 1
    end_col = 3
    ws.merge_cells(
        start_row=neighborhoods_row_number,
        start_column=start_col,
        end_row=neighborhoods_row_number,
        end_column=end_col,
    )
    cell = ws.cell(row=neighborhoods_row_number, column=start_col)
    cell.alignment = Alignment(wrap_text=True, horizontal="left", vertical="top")

    # Merged cells don't adjust height automatically, so we need to estimate it.
    if cell.value:
        merged_width = sum(
            ws.column_dimensions[col[0].column_letter].width
            for col in ws.iter_cols(min_col=start_col, max_col=end_col)
        )
        char_width = 1.2
        lines = 0
        for line in str(cell.value).split():
            line_length = len(line) * char_width
            lines += line_length / merged_width

        ws.row_dimensions[neighborhoods_row_number].height = max(15, math.ceil(lines) * 15)

    return


@typechecked
def _append_extra_notes(ws: Worksheet, extra_notes: list[str]) -> None:
    """Append extra notes to the worksheet."""
    start_row = ws.max_row + 2
    for i, note in enumerate(extra_notes, start=start_row):
        cell = ws.cell(row=i, column=5, value=note)
        cell.alignment = Alignment(wrap_text=True, horizontal="left")

    return



"""Utility functions for the bfb_delivery module."""

from datetime import datetime

import pandas as pd
from typeguard import typechecked


@typechecked
def get_friday(fmt: str) -> str:
    """Get the soonest Friday."""
    friday = datetime.now() + pd.DateOffset(weekday=4)
    return friday.strftime(fmt)



"""Field checks."""

import pandas as pd
import pandera.extensions as extensions

from bfb_delivery.lib.constants import CircuitColumns


@extensions.register_check_method(statistics=["flag"])
def at_least_two_words(pandas_obj: pd.Series, flag: bool) -> bool:
    """Check that a string has at least two words."""
    return all(len(val.split(" ")) >= 2 for val in pandas_obj) if flag else True


@extensions.register_check_method(statistics=["start_idx"])
def contiguous(pandas_obj: pd.Series, start_idx: int) -> bool:
    """Assert that values are contiguous."""
    return sorted(pandas_obj.to_list()) == list(
        range(start_idx, len(pandas_obj.to_list()) + start_idx)
    )


@extensions.register_check_method(statistics=["category_list"])
def in_list_case_insensitive(pandas_obj: pd.Series, *, category_list: list[str]) -> bool:
    """Check that a column is in a list."""
    return pandas_obj.str.upper().isin([val.upper() for val in category_list]).all()


@extensions.register_check_method(statistics=["flag"])
def is_list_of_one_or_less(pandas_obj: pd.Series, flag: bool) -> bool:
    """Check that a column is a list of one item."""
    return (
        all(isinstance(val, list) and 0 <= len(val) <= 1 for val in pandas_obj)
        if flag
        else True
    )


@extensions.register_check_method(statistics=["check_sort"])
def is_sorted(pandas_obj: pd.Series, check_sort: int) -> bool:
    """Assert that values are contiguous."""
    return sorted(pandas_obj.to_list()) == pandas_obj.to_list() if check_sort else True


@extensions.register_check_method(statistics=["item_name"])
def item_in_field_dict(pandas_obj: pd.Series, item_name: str) -> bool:
    """Check that a dictionary field has an item in it."""
    return all(item_name in val.keys() for val in pandas_obj)


@extensions.register_check_method(statistics=["flag"])
def one_product(pandas_obj: pd.Series, flag: bool) -> bool:
    """Ensure one and only one product per stop."""
    return (
        all(
            isinstance(val.get(CircuitColumns.PRODUCTS), list)
            and len(val.get(CircuitColumns.PRODUCTS)) == 1
            for val in pandas_obj
        )
        if flag
        else True
    )



"""Top-level init."""

from importlib.metadata import version

from bfb_delivery.api import (
    build_routes_from_chunked,
    combine_route_tables,
    create_manifests,
    create_manifests_from_circuit,
    format_combined_routes,
    split_chunked_route,
)

try:
    __version__: str = version(__name__)
except Exception:
    __version__ = "unknown"

del version



"""Schema checks."""

from bfb_delivery.lib.schema.checks.dataframe_checks import (
    address1_in_address,
    address2_in_address,
    at_least_one_in_group,
    at_least_one_in_group_sheet_plan,
    at_least_one_in_group_sheet_stop,
    contiguous_group,
    equal_cols,
    increasing_by,
    item_in_dict_col,
    many_to_one,
    one_to_one,
    one_to_one_route_sheet,
    unique_group,
)
from bfb_delivery.lib.schema.checks.field_checks import (
    at_least_two_words,
    contiguous,
    in_list_case_insensitive,
    is_list_of_one_or_less,
    is_sorted,
    item_in_field_dict,
)



"""Public and internal APIs."""

from bfb_delivery.api.public import (
    build_routes_from_chunked,
    combine_route_tables,
    create_manifests,
    create_manifests_from_circuit,
    format_combined_routes,
    split_chunked_route,
)



# noqa: D100
__doc__ = """
.. click:: bfb_delivery.cli.create_manifests_from_circuit:main
    :prog: create_manifests_from_circuit
    :nested: full
"""

import logging

import click
from typeguard import typechecked

from bfb_delivery import create_manifests_from_circuit
from bfb_delivery.lib.constants import Defaults, DocStrings

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


@click.command(help=DocStrings.CREATE_MANIFESTS_FROM_CIRCUIT.cli_docstring)
@click.option(
    "--start_date",
    type=str,
    required=False,
    default=Defaults.CREATE_MANIFESTS_FROM_CIRCUIT["start_date"],
    help=DocStrings.CREATE_MANIFESTS_FROM_CIRCUIT.args["start_date"],
)
@click.option(
    "--end_date",
    type=str,
    required=False,
    default=Defaults.CREATE_MANIFESTS_FROM_CIRCUIT["end_date"],
    help=DocStrings.CREATE_MANIFESTS_FROM_CIRCUIT.args["end_date"],
)
@click.option(
    "--output_dir",
    type=str,
    required=False,
    default=Defaults.CREATE_MANIFESTS_FROM_CIRCUIT["output_dir"],
    help=DocStrings.CREATE_MANIFESTS_FROM_CIRCUIT.args["output_dir"],
)
@click.option(
    "--output_filename",
    type=str,
    required=False,
    default=Defaults.CREATE_MANIFESTS_FROM_CIRCUIT["output_filename"],
    help=DocStrings.CREATE_MANIFESTS_FROM_CIRCUIT.args["output_filename"],
)
# TODO: Change circuit_output_dir default to subdir in output_dir.
@click.option(
    "--circuit_output_dir",
    type=str,
    required=False,
    default=Defaults.CREATE_MANIFESTS_FROM_CIRCUIT["circuit_output_dir"],
    help=DocStrings.CREATE_MANIFESTS_FROM_CIRCUIT.args["circuit_output_dir"],
)
@click.option(
    "--all_hhs",
    is_flag=True,
    default=Defaults.CREATE_MANIFESTS_FROM_CIRCUIT["all_hhs"],
    help=DocStrings.CREATE_MANIFESTS_FROM_CIRCUIT.args["all_hhs"],
)
@click.option(
    "--verbose",
    is_flag=True,
    default=Defaults.CREATE_MANIFESTS_FROM_CIRCUIT["verbose"],
    help="verbose: Flag to print verbose output.",
)
@click.option(
    "--extra_notes_file",
    type=str,
    required=False,
    default=Defaults.CREATE_MANIFESTS_FROM_CIRCUIT["extra_notes_file"],
    help=DocStrings.CREATE_MANIFESTS_FROM_CIRCUIT.args["extra_notes_file"],
)
@typechecked
def main(  # noqa: D103
    start_date: str,
    end_date: str,
    output_dir: str,
    output_filename: str,
    circuit_output_dir: str,
    all_hhs: bool,
    verbose: bool,
    extra_notes_file: str,
) -> str:
    final_manifest_path, new_circuit_output_dir = create_manifests_from_circuit(
        start_date=start_date,
        end_date=end_date,
        output_dir=output_dir,
        output_filename=output_filename,
        circuit_output_dir=circuit_output_dir,
        all_hhs=all_hhs,
        verbose=verbose,
        extra_notes_file=extra_notes_file,
    )
    logger.info(f"Formatted workbook saved to:\n{final_manifest_path.resolve()}")
    # Print statement to capture in tests.
    print(str(final_manifest_path))
    print(str(new_circuit_output_dir))

    return str(final_manifest_path)
